/*
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree.
 */

#include "LoggingAPI.h"

#include "Logger.h"

namespace libkineto {
int getLogSeverityLevel() {
  return Logger::severityLevel();
}

void setLogSeverityLevel(int level) {
  SET_LOG_SEVERITY_LEVEL(level);
}
} // namespace libkineto
